use crate::schema;
use microrm::prelude::*;
use serde::{Deserialize, Serialize};

mod helper;

fn default_auth_token_expiry() -> u64 {
    600
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Config {
    pub base_url: String,

    #[serde(default = "default_auth_token_expiry")]
    pub auth_token_expiry: u64,
}

impl Config {
    pub fn build_from(db: &schema::UIDCDatabase, cfile: Option<&str>) -> Self {
        let mut config_map = std::collections::HashMap::<String, String>::new();
        // load config keys from database
        let db_pcs = db.persistent_config.get().expect("could't get config keys from database");
        config_map.extend(db_pcs.into_iter().map(|pc: microrm::schema::IDWrap<schema::PersistentConfig>| {
            let pc = pc.wrapped();
            (pc.key, pc.value)
        }));

        if let Some(path) = cfile {
            match std::fs::read(&path) {
                Ok(data) => {
                    log::info!("Loading config from {path}...");
                    let toml_table: toml::Table = toml::from_str(
                        std::str::from_utf8(data.as_slice())
                            .expect("couldn't read config file contents as utf-8"),
                    )
                    .expect("couldn't parse config toml");

                    for val in toml_table {
                        log::trace!("using config key {} from TOML config...", val.0);
                        if val.1.is_str() {
                            config_map.insert(val.0, val.1.as_str().unwrap().to_string());
                        } else {
                            config_map.insert(val.0, val.1.to_string());
                        }
                    }
                }
                Err(e) => {
                    log::error!("Could not open {path} for reading: {e}");
                }
            }
        }

        let mut deser = helper::ConfigDeserializer {
            config_map: &config_map,
            prefix: "".to_string(),
        };

        let config = Config::deserialize(&mut deser).expect("couldn't load configuration");

        log::trace!("final configuration: {:?}", config);

        config
    }

    pub fn save<'config, 'qi>(&'config self, db: &'config schema::UIDCDatabase) {
        let ser = helper::ConfigSerializer {
            config: &self,
            db,
            prefix: String::new(),
        };

        let _ = self.serialize(&ser);
    }
}
