use crate::key::ParsedKey;

fn jwt_algorithm(pk: &ParsedKey) -> &'static str {
    match pk {
        ParsedKey::Ed25519 { key_id: _, keypair: _ } => "EdDSA",
        ParsedKey::RSA { key_id: _, keypair: _ } => "RS256",
    }
}

#[derive(serde::Serialize, serde::Deserialize)]
pub struct JWTHeader {
    pub alg: Option<String>,
    pub typ: String,
    pub kid: Option<String>,
}

impl JWTHeader {
    pub fn parse(full_jwt: &str) -> Option<Self> {
        let header_raw = full_jwt.split(".").next()?;
        let header_decoded = base64::decode_config(header_raw.as_bytes(), base64::URL_SAFE_NO_PAD).ok()?;

        serde_json::from_slice(header_decoded.as_slice()).ok()
    }
}

#[derive(serde::Serialize, serde::Deserialize)]
pub struct JWTData<'l> {
    pub sub: &'l str,
    pub iss: &'l str,
    pub aud: &'l str,
    pub iat: u64,
    pub exp: u64,

    #[serde(flatten)]
    pub extras: std::collections::HashMap<&'l str, serde_json::Value>,
}

impl<'l> Into<String> for JWTData<'l> {
    fn into(self) -> String {
        serde_json::to_string(&self).expect("JSON serialization failure?")
    }
}

pub struct JWT {
    pub header: String,
    pub unencoded_data: Vec<u8>,
    pub encoded_data: String,
    pub signature: String,
}

impl JWT {
    pub fn verify(
        with: &ParsedKey,
        from: &str,
    ) -> Option<Self> {
        let header_split = from.find(".")?;
        let header = &from[0..header_split];
        let data_split = header_split + 1 + from[header_split + 1..].find(".")?;
        let data = &from[header_split + 1..data_split];
        let signature = &from[data_split + 1..];

        let mut to_verify = vec![];
        to_verify.extend(header.as_bytes());
        to_verify.extend(".".as_bytes());
        to_verify.extend(data.as_bytes());

        let decoded_signature =
            base64::decode_config(signature.as_bytes(), base64::URL_SAFE_NO_PAD).ok()?;
        with.verify_signature(to_verify.as_ref(), decoded_signature.as_ref())
            .ok()?;

        // if we got this far, the verification passed
        Some(Self {
            header: header.into(),
            unencoded_data: base64::decode_config(data.as_bytes(), base64::URL_SAFE_NO_PAD).ok()?,
            encoded_data: data.to_string(),
            signature: signature.into(),
        })
    }

    pub fn claims(&self) -> Option<JWTData> {
        serde_json::from_slice(self.unencoded_data.as_slice()).ok()
    }

    pub fn sign(with: &ParsedKey, data: JWTData) -> Self {
        let header = JWTHeader {
            alg: Some(jwt_algorithm(with).into()),
            typ: "JWT".into(),
            kid: Some(with.key_id().into()),
        };
        let header_data = base64::encode_config(serde_json::to_vec(&header).unwrap(), base64::URL_SAFE_NO_PAD);

        let unencoded_data = Into::<String>::into(data);
        let data = base64::encode_config(unencoded_data.as_bytes(), base64::URL_SAFE_NO_PAD);

        let mut to_sign = vec![];
        to_sign.extend(header_data.as_bytes());
        to_sign.extend(".".as_bytes());
        to_sign.extend(data.as_bytes());
        let signature =
            base64::encode_config(with.generate_signature(&to_sign).expect("couldn't sign data"), base64::URL_SAFE_NO_PAD);

        Self {
            header: header_data,
            unencoded_data: unencoded_data.as_bytes().into(),
            encoded_data: data,
            signature,
        }
    }

    pub fn into_string(self) -> String {
        self.header + "." + self.encoded_data.as_str() + "." + self.signature.as_str()
    }
}

#[cfg(test)]
mod test {
    use ring::signature::KeyPair;

    #[test]
    fn simple_round_trip() {
        let rng = ring::rand::SystemRandom::new();
        let kpair_raw = ring::signature::Ed25519KeyPair::generate_pkcs8(&rng)
            .expect("couldn't generate ephemeral keypair");
        let kpair = ring::signature::Ed25519KeyPair::from_pkcs8(kpair_raw.as_ref())
            .expect("couldn't load keypair");

        let jdata = super::JWTData {
            sub: "sub",
            iss: "iss",
            aud: "aud",
            iat: 1,
            exp: 2,
            extras: Default::default(),
        };

        let generated = super::JWT::sign(&kpair, jdata).into_string();
        println!("generated: {:?}", generated);

        let pubkey = ring::signature::UnparsedPublicKey::new(
            &ring::signature::ED25519,
            kpair.public_key().as_ref(),
        );

        let vresult = super::JWT::verify(&pubkey, generated.as_str());

        assert!(vresult.is_some());
    }
}
