use std::{cell::RefCell, sync::Arc};

use crate::{schema, UIDCError};
use ring::signature::{Ed25519KeyPair, KeyPair};
use sha2::Digest;
use microrm::prelude::*;

#[derive(Debug)]
pub enum KeyError {
    Plain(&'static str),
}

#[non_exhaustive]
#[derive(Clone, Copy, Debug, PartialEq, serde::Serialize, serde::Deserialize)]
pub enum KeyType {
    RSA2048,
    RSA4096,
    Ed25519,
}

impl std::fmt::Display for KeyType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        <Self as std::fmt::Debug>::fmt(self, f)
    }
}

pub const KEY_TYPE_NAMES : &'static [(&'static str, KeyType)] = &[
    ("rsa2048", KeyType::RSA2048),
    ("rsa4096", KeyType::RSA4096),
    ("ed25519", KeyType::Ed25519),
];

impl std::str::FromStr for KeyType {
    type Err = UIDCError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        for (spec, kty) in KEY_TYPE_NAMES {
            if s == *spec {
                return Ok(*kty)
            }
        }
        Err(UIDCError::Abort("invalid keytype"))
    }
}

pub enum ParsedKey {
    Ed25519 { key_id: String, keypair: ring::signature::Ed25519KeyPair },
    RSA { key_id: String, keypair: ring::signature::RsaKeyPair }
}

impl ParsedKey {
    pub fn key_id(&self) -> &str {
        match self {
            Self::Ed25519 { key_id, keypair } => key_id.as_str(),
            Self::RSA { key_id, keypair } => key_id.as_str(),
        }
    }

    pub fn generate_signature(&self, data: &[u8]) -> Result<Vec<u8>, UIDCError> {
        match self {
            Self::Ed25519 { key_id, keypair } => {
                Ok(keypair.sign(data).as_ref().into())
            },
            Self::RSA { key_id, keypair } => {
                let rng = ring::rand::SystemRandom::new();
                let mut signature = vec![];
                signature.resize(keypair.public_modulus_len(), 0);
                keypair.sign(&ring::signature::RSA_PKCS1_SHA256, &rng, data, signature.as_mut_slice()).map_err(|_| KeyError::Plain("failed to generate RSA signature!"))?;
                Ok(signature)
            }
        }
    }

    pub fn verify_signature(&self, data: &[u8], signature: &[u8]) -> Result<bool, UIDCError> {
        use ring::signature::VerificationAlgorithm;
        match self {
            Self::Ed25519 { keypair, .. } => {
                Ok(ring::signature::ED25519.verify(keypair.public_key().as_ref().into(), data.into(), signature.into()).is_ok())
            },
            Self::RSA { keypair, .. } => {
                Ok(ring::signature::RSA_PKCS1_2048_8192_SHA256.verify(keypair.public_key().as_ref().into(), data.into(), signature.into()).is_ok())
            },
        }
    }
}

pub struct RealmKeys {
    realm: schema::Realm,
    // key cache, to avoid reparsing the key data repeatedly
    keys: RefCell<Vec<(KeyType, Arc<ParsedKey>)>>,
}

impl RealmKeys {
    pub fn new(realm: schema::Realm) -> Self {
        Self { realm, keys: vec![].into() }
    }

    pub fn by_key_id(&mut self, id: &String) -> Result<Option<Arc<ParsedKey>>, UIDCError> {
        // check the cache
        for key in self.keys.borrow().iter() {
            if key.1.key_id() == id {
                return Ok(Some(key.1.clone()))
            }
        }

        // then check the database
        let key = self.realm.keys.with(schema::Key::KeyId, id).first().get()?;

        if let Some(key) = key {
            let parsed = Arc::new(match key.key_type.as_ref() {
                KeyType::RSA2048
                | KeyType::RSA4096 => {
                    todo!()
                },
                KeyType::Ed25519 => {
                    ParsedKey::Ed25519 { key_id: key.key_id.clone(), keypair: Ed25519KeyPair::from_pkcs8(&key.secret_data).map_err(|_| UIDCError::Abort("could not load ed25519 key from database"))? }
                },
            });

            self.keys.borrow_mut().push((*key.key_type.as_ref(), parsed.clone()));

            Ok(Some(parsed))
        }
        else {
            Ok(None)
        }
    }

    pub fn by_type(&mut self, kty: KeyType) -> Result<Option<Arc<ParsedKey>>, UIDCError> {
        // check the cache
        for key in self.keys.borrow().iter() {
            if key.0 == kty {
                return Ok(Some(key.1.clone()))
            }
        }

        // then check the database
        let key = self.realm.keys.with(schema::Key::KeyType, &kty.into()).first().get()?;

        if let Some(key) = key {
            let parsed = Arc::new(match kty {
                KeyType::RSA2048
                | KeyType::RSA4096 => {
                    ParsedKey::RSA { key_id: key.key_id.clone(), keypair: ring::signature::RsaKeyPair::from_pkcs8(&key.secret_data).map_err(|_| UIDCError::Abort("could not load RSA key from database"))? }
                },
                KeyType::Ed25519 => {
                    ParsedKey::Ed25519 { key_id: key.key_id.clone(), keypair: Ed25519KeyPair::from_pkcs8(&key.secret_data).map_err(|_| UIDCError::Abort("could not load ed25519 key from database"))? }
                },
            });

            self.keys.borrow_mut().push((kty, parsed.clone()));

            Ok(Some(parsed))
        }
        else {
            Ok(None)
        }
    }
}

fn pubkey_id(data: &[u8]) -> String {
    let mut key_hasher = sha2::Sha256::new();
    key_hasher.update(data);
    let mut key_id = base64::encode(key_hasher.finalize());
    key_id.truncate(16);
    key_id
}

fn generate_rsa(realm: &schema::Realm, kty: KeyType, bits: usize) -> Result<ParsedKey, UIDCError> {
    // ring does not generate RSA keypairs, so we need to shell out to openssl for this.
    let openssl_output = std::process::Command::new("sh")
        .arg("-c")
        .arg(format!(
            "openssl genpkey \
             -algorithm RSA \
             -pkeyopt rsa_keygen_pubexp:65537 \
             -pkeyopt rsa_keygen_bits:{bits} \
             | openssl pkcs8 \
             -topk8 \
             -nocrypt \
             -outform der"))
        .output()
        .map_err(|_| UIDCError::Abort("couldn't invoke openssl"))?;

    let secret = openssl_output.stdout;

    let keypair = ring::signature::RsaKeyPair::from_pkcs8(&secret).map_err(|_| UIDCError::Abort("couldn't parse PKCS#8 output from openssl"))?;
    let public = keypair.public_key().as_ref();
    let key_id = pubkey_id(public.as_ref());
    let expiry = time::OffsetDateTime::now_utc() + time::Duration::days(730);

    realm.keys.insert(schema::Key {
        key_id: key_id.clone(),
        key_type: kty.into(),
        public_data: public.into(),
        secret_data: secret.into(),
        expiry
    })?;

    Ok(ParsedKey::RSA { key_id, keypair })
}

pub fn generate_in(realm: &schema::Realm, kty: KeyType) -> Result<ParsedKey, UIDCError> {
    let mut rng = ring::rand::SystemRandom::new();
    match kty {
        KeyType::RSA2048 => {
            generate_rsa(realm, KeyType::RSA2048, 2048)
        },
        KeyType::RSA4096 => {
            generate_rsa(realm, KeyType::RSA4096, 4096)
        },
        KeyType::Ed25519 => {
            let generated_keypair = Ed25519KeyPair::generate_pkcs8(&mut rng)
                .map_err(|_| KeyError::Plain("failed to generate key"))?;

            let keydata = generated_keypair.as_ref().to_owned();

            let loaded_key = Ed25519KeyPair::from_pkcs8(keydata.as_slice())
                .expect("couldn't load just-generated key");
            let pubkey = loaded_key.public_key();
            let key_id = pubkey_id(pubkey.as_ref());
            let expiry = time::OffsetDateTime::now_utc() + time::Duration::days(730);

            realm.keys.insert(schema::Key {
                key_id: key_id.clone(),
                key_type: kty.into(),
                // no separate public data for EdDSA keys
                public_data: vec![],
                secret_data: keydata,
                expiry
            })?;

            Ok(ParsedKey::Ed25519 { key_id, keypair: loaded_key })
        },
    }
}

pub fn list(realm: &schema::Realm) -> Result<(), UIDCError> {
    let keys = realm.keys.get()?;

    for key in keys {
        println!("- [{}] {:?}, expires {}", key.key_id, key.key_type, key.expiry.format(&time::format_description::well_known::Rfc3339).unwrap());
    }
    Ok(())
}

pub fn remove(realm: &schema::Realm, key_id: &String) -> Result<(), UIDCError> {
    realm.keys.with(schema::Key::KeyId, key_id).delete()?;
    Ok(())
}
