use crate::{config::Config, schema, token, UIDCError, jwt, key};
use microrm::prelude::*;

pub fn create_auth_token(
    realm: &schema::Realm,
    config: &Config,
    client: &String,
    username: &String,
    scopes: &String,
) -> Result<String, UIDCError> {
    token::generate_auth_token(
        config,
        realm,
        &realm.clients.unique(client).get()?
            .ok_or(UIDCError::Abort("no such client"))?.wrapped(),
        &realm.users.unique(username).get()?
            .ok_or(UIDCError::Abort("no such user"))?.wrapped(),
        scopes.split_whitespace(),
    )
}

pub fn create_refresh_token(
    realm: &schema::Realm,
    config: &Config,
    client: &String,
    username: &String,
    scopes: &String,
) -> Result<String, UIDCError> {
    token::generate_refresh_token(
        config,
        realm,
        &realm.clients.unique(client).get()?
            .ok_or(UIDCError::Abort("no such client"))?.wrapped(),
        &realm.users.unique(username).get()?
            .ok_or(UIDCError::Abort("no such user"))?.wrapped(),
        scopes.split_whitespace(),
    )
}

pub fn inspect_token(
    config: &Config,
    realm: &schema::Realm,
    token: Option<&String>,
) -> Result<(), UIDCError> {
    let token = match token {
        Some(token) => token.clone(),
        None => rpassword::prompt_password("Enter token: ").unwrap(),
    };

    let header = jwt::JWTHeader::parse(token.as_str()).ok_or(UIDCError::Abort("failed to parse JWT header"))?;

    let mut realmkeys = key::RealmKeys::new(realm.clone());

    let mut pk = None;

    if let Some(kid) = header.kid {
        pk = realmkeys.by_key_id(&kid)?;
    }
    if pk.is_none() {
        if let Some(_alg) = header.alg {
            todo!("algorithm fallback")
            // pk = realmkeys.by_type(&kid)?;
        }
    }

    let jwt = if let Some(key) = pk {
        jwt::JWT::verify(&key, token.as_str())
    }
    else {
        println!("No matching realm key found!");
        return Ok(());
    };

    // let jwt = jwt::JWT::verify(&pubkey, token.as_str());
    if jwt.is_none() {
        println!("Signature validation against realm key failed!");
    } else if let Some(claims) = jwt.as_ref().and_then(jwt::JWT::claims) {
        println!("Base claims:");
        println!(" - issuer      : {}", claims.iss);
        println!(" - audience    : {}", claims.aud);
        println!(" - subject     : {}", claims.sub);
        println!(
            " - issued at   : {} [{}]",
            claims.iat,
            time::OffsetDateTime::from_unix_timestamp(claims.iat as i64).unwrap()
        );
        println!(
            " - expires at  : {} [{}]",
            claims.exp,
            time::OffsetDateTime::from_unix_timestamp(claims.exp as i64).unwrap()
        );
        for claim in claims.extras {
            println!(" - {:12}: {}", claim.0, claim.1);
        }
    } else {
        println!("Claim parsing failed!");
    }

    Ok(())
}
