use crate::{schema, UIDCError};
use microrm::prelude::*;

pub fn list(realm: &schema::Realm) -> Result<(), UIDCError> {
    let users = realm.users.get()?;

    println!("User list ({} users):", users.len());

    for user in &users {
        println!("- {:20}", user.username);
        let auth_challenges = user.auth.get()?;
        for ch in &auth_challenges {
            println!("    - Has {:?} authentication challenge", ch.challenge_type);
        }
    }

    Ok(())
}

pub fn create(
    realm: &schema::Realm,
    username: &String,
) -> Result<(), UIDCError> {
    // check that the user doesn't exist already
    let existing_user = realm.users.unique(&username).get()?;

    if existing_user.is_some() {
        log::error!(
            "Can't create user {} in realm, as a user with that username already exists",
            username,
        );
        return Ok(());
    }

    realm.users.insert(schema::User {
        username: username.into(),
        auth: Default::default(),
        groups: Default::default(),
    })?;

    Ok(())
}

pub fn change_auth(
    realm: &schema::Realm,
    username: &String,
    change_password: bool,
    change_totp: bool,
) -> Result<(), UIDCError> {
    // check that the user exists
    let user = realm.users.unique(username).get()?.ok_or(UIDCError::Abort("no such user"))?;

    let user = crate::user::User::from_schema(realm, user);

    if change_password {
        let raw_pass = rpassword::prompt_password("Enter new user password: ").unwrap();
        user.set_new_password(raw_pass.as_bytes())?;
    }
    if change_totp {
        let (new_secret, new_uri) = user.generate_totp_with_uri()?;
        println!("Please confirm you can generate tokens with the new secret:");
        qr2term::print_qr(new_uri.as_str())
            .map_err(|_| UIDCError::Abort("could not display QR code"))?;
        let new_challenge = schema::AuthChallenge {
            challenge_type: schema::AuthChallengeType::TOTP.into(),
            public: vec![],
            secret: new_secret.clone(),
            enabled: true,
        };

        loop {
            let digits = rpassword::prompt_password("TOTP code: ").unwrap();
            if new_challenge.verify_totp_challenge(digits.as_bytes())? {
                break;
            }
        }
        user.set_new_totp(new_secret.as_slice())?;
    }
    Ok(())
}

pub fn inspect(
    realm: &schema::Realm,
    username: &String,
) -> Result<(), UIDCError> {
    let user = realm.users.unique(username).get()?;

    if let Some(user) = user {
        println!("User found: {}", username);
        println!("Groups:");
        for group in user.groups.get()? {
            println!(" - {}", group.shortname);
        }

        println!("Authentication methods:");

        for challenge in user.auth.get()? {
            println!(" - {:?}", challenge.challenge_type);
        }
    } else {
        println!("No such user {} in realm", username);
    }

    Ok(())
}
